/******************************************************************************
*
*       IrqMon - Interrupt Monitor for Windows 95
*		
*       03/17/97 SSS Adopted from Regmon
*
*       GUI Adopted from "RegMon" application
*		By Mark Russinovich and Bryce Cogswell
*		WDJ October 1996
*
*       FILE : Irqmon.c
*
*       PURPOSE : Communicates with the IrqMon driver 
*                 to display interrupt activity information.
*
******************************************************************************/

#include <windows.h>    // includes basic windows functionality
#include <windowsx.h>
#include <commctrl.h>   // includes the common control header
#include <stdio.h>
#include <string.h>
#include "resource.h"
#include "..\\vxds\\irqhook\\ioctl.h"

#define				VXD_FILE		"\\\\.\\IRQHOOK.VXD"

#define				VXD_NAME		"IRQHOOK"
#define				HELPER_NAME		"HELPER.VXD"
static HANDLE		VxDHandle		= INVALID_HANDLE_VALUE;
#define				TIMER_MS        500

// Buffer into which driver can copy statistics
DWORD				Stats[ STORE_SIZE_DWORDS ];
// Current fraction of buffer filled
DWORD				StatsLen;

// Application instance handle
HINSTANCE			hInst;

// Misc globals
HWND				hWndList;
BOOLEAN				Capture = TRUE;
BOOLEAN				Autoscroll = TRUE;
BOOLEAN				Deleting = FALSE;

// A bitmask of initially monitored IRQs (all by default)
DWORD				IRQsToMonitor=0xFFFF;  
// Hookable IRQs ; to be gotten from the VxD
DWORD				Hookable_IRQs;

// General buffer for storing temporary strings
static TCHAR		msgbuf[ 257 ];
struct TIMESTAMP	TimeStamp;

// General cursor manipulation
HCURSOR 			hSaveCursor;
HCURSOR 			hHourGlass;

// procs
long APIENTRY 		MainWndProc( HWND, UINT, UINT, LONG );
BOOL APIENTRY       About( HWND, UINT, UINT, LONG );

//functions
BOOL 				InitApplication( HANDLE );
HWND 				InitInstance( HANDLE, int );
HWND 				CreateList( HWND );
void 				UpdateStatistics( HWND hWnd, HWND hWndList, BOOL Clear );
void				InstallHelperAndReboot( HWND hWnd );

/****************************************************************************
*
*	FUNCTION:	Abort:
*
*	PURPOSE:	Handles emergency exit conditions.
*
*****************************************************************************/
void Abort( HWND hWnd, TCHAR * Msg )
{
	MessageBox( hWnd, Msg, "IrqMon", MB_OK );
	PostQuitMessage( 1 );
}		

/****************************************************************************
*
*	FUNCTION: WinMain(HANDLE, HANDLE, LPSTR, int)
*
*	PURPOSE:	calls initialization function, processes message loop
*
****************************************************************************/
int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
						LPSTR lpCmdLine, int nCmdShow )
{
	MSG 	msg;      
	HWND	hWnd;
        
	if ( ! InitApplication( hInstance ) )
		return FALSE;     

	// initializations that apply to a specific instance 
	if ( (hWnd = InitInstance( hInstance, nCmdShow )) == NULL )
		return FALSE;


	// acquire and dispatch messages until a WM_QUIT message is received.
	while ( GetMessage( &msg, NULL, 0, 0 ) )  {
		TranslateMessage( &msg );
		DispatchMessage( &msg ); 
	}
	return msg.wParam;										 
}

/****************************************************************************
*
*    FUNCTION: InitApplication(HANDLE)
*
*    PURPOSE: Initializes window data and registers window class
*
****************************************************************************/
BOOL InitApplication( HANDLE hInstance )
{
	WNDCLASS  wc;
	
	// Fill in window class structure with parameters that describe the
	// main (statistics) window. 
	wc.style			= 0;                     
	wc.lpfnWndProc		= (WNDPROC)MainWndProc; 
	wc.cbClsExtra		= 0;              
	wc.cbWndExtra		= 0;              
	wc.hInstance		= hInstance;       
	wc.hIcon			= LoadIcon( hInstance, "ICON" );
	wc.hCursor			= LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground	= GetStockObject( LTGRAY_BRUSH ); 
	wc.lpszMenuName		= "LISTMENU";  
	wc.lpszClassName	= "IrqMonClass";
	if ( ! RegisterClass( &wc ) )
		return FALSE;

	return TRUE;
}

/****************************************************************************
*
*    FUNCTION:  InitInstance(HANDLE, int)
*
*    PURPOSE:  Saves instance handle and creates main window
*
****************************************************************************/
HWND InitInstance( HANDLE hInstance, int nCmdShow )
{
	HWND hWndMain;

	hInst = hInstance;

	hWndMain = CreateWindow( "IrqMonClass", "Windows95 Interrupt Monitor", 
							WS_OVERLAPPEDWINDOW,
							CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
							NULL, NULL, hInstance, NULL );

	// if window could not be created, return "failure" 
	if ( ! hWndMain )
		return NULL;
	
	// make the window visible; update its client area; and return "success"
	ShowWindow( hWndMain, nCmdShow );
	UpdateWindow( hWndMain ); 
	return hWndMain;      
}

/****************************************************************************
*
*    FUNCTION: MainWndProc(HWND, unsigned, WORD, LONG)
*
*    PURPOSE:  Processes messages for the statistics window.
*
****************************************************************************/
LONG APIENTRY MainWndProc( HWND hWnd, UINT message, UINT wParam, LONG lParam) 
{
	DWORD			nb;
	TCHAR			Path[ 256 ];
	HDC				hDC;
	PAINTSTRUCT		Paint;
    UINT			i;

	switch ( message ) {

		case WM_CREATE:

			// get hourglass cursor ready
			hHourGlass = LoadCursor( NULL, IDC_WAIT );

			// post hourglass icon
			SetCapture(hWnd);
			hSaveCursor = SetCursor(hHourGlass);

			// Create the ListBox within the main window
			hWndList = CreateList( hWnd );
			if ( hWndList == NULL )
				MessageBox( NULL, "List not created!", NULL, MB_OK );

		    // open the handle to the device
			GetCurrentDirectory( sizeof Path, Path );
			wsprintf( Path+lstrlen(Path), "\\%s", VXD_FILE );
			// connect with VxD
			//VxDHandle = CreateFile( Path, 0, 0, NULL,
			VxDHandle = CreateFile( VXD_FILE, 0, 0, NULL,
							0, FILE_FLAG_OVERLAPPED|
							FILE_FLAG_DELETE_ON_CLOSE,
							NULL );
			if ( VxDHandle == INVALID_HANDLE_VALUE )  {
				wsprintf( msgbuf, "%s is not loaded properly.", VXD_NAME );
				Abort( hWnd, msgbuf );
				return FALSE;
			}

			// Have driver zero information
			if ( ! DeviceIoControl(	VxDHandle, CMD_reset,
									NULL, 0, NULL, 0, &nb, NULL ) )
			{
				Abort( hWnd, "Couldn't access device driver") ;
				return TRUE;
			}

			// Get IRQs which were hooked successfully
			if ( ! DeviceIoControl(	VxDHandle, CMD_get_hookable_irq_bitmask,
									NULL, 0, 
									&Hookable_IRQs, sizeof Hookable_IRQs, 
									&nb, NULL ) )
			{
				Abort( hWnd, "Couldn't access device driver") ;
				return TRUE;
			}

			if ((IRQsToMonitor & ~Hookable_IRQs) & 0xFFFF)
					InstallHelperAndReboot(hWnd);

			// Filter out non-monitorable IRQs (if Helper.VxD is not installed)
			IRQsToMonitor &= Hookable_IRQs;

			// Tell VxD what irqs to monitor
			if ( ! DeviceIoControl(	VxDHandle, CMD_setirqs,
									&IRQsToMonitor, sizeof IRQsToMonitor, 
									NULL, 0, 
									&nb, NULL ) )
			{
				Abort( hWnd, "Couldn't access device driver") ;
				return TRUE;
			}
			
			for (i=0; i<15; i++) {

				EnableMenuItem( GetMenu(hWnd), i+IDM_IRQ0, MF_BYCOMMAND | 
							   Hookable_IRQs & (1<<i) ?
							   MF_ENABLED : MF_GRAYED ); 
				CheckMenuItem ( GetMenu(hWnd), i+IDM_IRQ0, MF_BYCOMMAND | 
					           IRQsToMonitor & (1<<i) ?
							   MF_CHECKED : MF_UNCHECKED ); 
				}

			// Start up timer to periodically update screen
			SetTimer( hWnd,	1, TIMER_MS/*ms*/, NULL );
			
			// Initialization done
			SetCursor( hSaveCursor );
			ReleaseCapture();
			return 0;

		case WM_NOTIFY:
			// Make sure its intended for us
			if ( wParam == ID_LIST )  {
				NM_LISTVIEW	* pNm = (NM_LISTVIEW *)lParam;
				switch ( pNm->hdr.code )  {

			        case LVN_BEGINLABELEDIT:
						// Don't allow editing of information
						return TRUE;
				}
			}
			return 0;

		case WM_COMMAND:

			switch ( LOWORD( wParam ) )	 {

				// stats related commands to send to driver
				case IDM_CLEAR:
					// Have driver zero information
					if ( ! DeviceIoControl(	VxDHandle, CMD_reset,
											NULL, 0, NULL, 0, &nb, NULL ) )
					{
						Abort( hWnd, "Couldn't access device driver");
						return TRUE;
					}
					// Update statistics windows
					UpdateStatistics( hWnd, hWndList, TRUE );
					return 0;

				case IDM_CAPTURE:
					// Read statistics from driver
					Capture = !Capture;
					CheckMenuItem( GetMenu(hWnd), IDM_CAPTURE,
									MF_BYCOMMAND|(Capture?MF_CHECKED:MF_UNCHECKED) ); 
					return 0;

				case IDM_AUTOSCROLL:
					Autoscroll = !Autoscroll;
					CheckMenuItem( GetMenu(hWnd), IDM_AUTOSCROLL,
									MF_BYCOMMAND|(Autoscroll?MF_CHECKED:MF_UNCHECKED) ); 
					return 0;

				case IDM_EXIT:
					// Close ourself
					SendMessage( hWnd, WM_CLOSE, 0, 0 );
					return 0;

				case IDM_ABOUT:
					// Show the names of the authors
					DialogBox( hInst, "AboutBox", hWnd, (DLGPROC)About );
					return 0;

				case IDM_IRQ0  : case IDM_IRQ1  : case IDM_IRQ2 : case IDM_IRQ3 :
				case IDM_IRQ4  : case IDM_IRQ5  : case IDM_IRQ6 : case IDM_IRQ7 :
				case IDM_IRQ8  : case IDM_IRQ9  : case IDM_IRQ10 : case IDM_IRQ11 :
				case IDM_IRQ12 : case IDM_IRQ13 : case IDM_IRQ14 : case IDM_IRQ15 :
					// Enable/Disable IRQ monitoring
					IRQsToMonitor ^= 1<<(wParam-IDM_IRQ0); 

					CheckMenuItem( GetMenu(hWnd), wParam,
									MF_BYCOMMAND| (IRQsToMonitor & 1<<(wParam-IDM_IRQ0)
									? MF_CHECKED:MF_UNCHECKED) ); 

					if ( ! DeviceIoControl( VxDHandle, CMD_setirqs,
										&IRQsToMonitor, sizeof IRQsToMonitor,
										NULL, 0,
										NULL, NULL ) )
					{
						Abort( hWnd, "Couldn't access device driver") ;
						return TRUE;
					}
					return 0;

				default:
					// Default behavior
					return DefWindowProc( hWnd, message, wParam, lParam );
			}
			break;

		case WM_TIMER:
			
			// Time to query the device driver for more data
			if (Capture)
			  {
					// Have driver fill Stats buffer with information
			  if ( ! DeviceIoControl( VxDHandle, CMD_getstat,
											NULL, 0, Stats, sizeof Stats,
											&StatsLen, NULL ) )
				{
					Abort( hWnd, "Couldn't access device driver" );
					return TRUE;
				}
				
			  if ( StatsLen == 0 )
				break;

			  // Update statistics windows
			  UpdateStatistics( hWnd, hWndList, /*TRUE*/ FALSE );
			  }

			return 0;

		case WM_SIZE:
			// Move or resize the List
            MoveWindow( hWndList, 0, 0, LOWORD(lParam), HIWORD(lParam), TRUE );
            return 0;

		case WM_CLOSE:
			//PostQuitMessage( 0 );
			CloseHandle( VxDHandle );
			return DefWindowProc( hWnd, message, wParam, lParam );

		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;

		case WM_PAINT:
			if( Deleting ) {
				hDC = BeginPaint( hWnd, &Paint );
				EndPaint( hWnd, &Paint );
				return 1;
			}
			return DefWindowProc( hWnd, message, wParam, lParam );

		default:
			// Default behavior
			return DefWindowProc( hWnd, message, wParam, lParam );
	}
	return 0;
}

/******************************************************************************
*
*	FUNCTION:	Split
*
*	PURPOSE:	Split a delimited line into components
*
******************************************************************************/
int Split( char * line, char delimiter, char * items[] )
{
	int		cnt = 0;

	for (;;)  {
		// Add prefix to list of components		
		items[cnt++] = line;

		// Check for more components
		line = strchr( line, delimiter );
		if ( line == NULL )
			return cnt;

		// Terminate previous component and move to next
		*line++ = '\0';
	}		
}

/******************************************************************************
*
*	FUNCTION:	ListAppend
*
*	PURPOSE:	Add a new line to List window
*
******************************************************************************/
BOOL List_Append( HWND hWndList, DWORD seq, char * line )
{
	LV_ITEM		lvI;	// list view item structure
	int			row;
	char	*	items[20];
	int			itemcnt = 0;

	// Split line into columns
	itemcnt = Split( line, '\t', items );
	if ( itemcnt == 0 )
		return TRUE;

	// Determine row number for request
	if ( *items[0] )  {
		// Its a new request.  Put at end.
		row = 0x7FFFFFFF;
	} else {
		// Its a status.  Locate its associated request.
		lvI.mask = LVIF_PARAM;
		lvI.iSubItem = 0;
		for ( row = ListView_GetItemCount(hWndList) - 1; row >= 0; --row )  {
			lvI.iItem = row;
			if ( ListView_GetItem( hWndList, &lvI )  &&  (DWORD)lvI.lParam == seq )
				break;
		}
		if ( row == -1 )
			// No request associated with status.
			return TRUE;
	}

	// Sequence number if a new item
	if ( *items[0] )  {
		wsprintf( msgbuf, "%d", seq );
		lvI.mask		= LVIF_TEXT | LVIF_PARAM;
		lvI.iItem		= row;
		lvI.iSubItem	= 0;
		lvI.pszText		= msgbuf;
		lvI.cchTextMax	= lstrlen( lvI.pszText ) + 1;
		lvI.lParam		= seq;
		row = ListView_InsertItem( hWndList, &lvI );
		if ( row == -1 )  {
			wsprintf( msgbuf, "Error adding item %d to list view", seq );
			MessageBox( hWndList, msgbuf, "IrqMon Error", MB_OK );
			return FALSE;
		}
	}

	// Insert IRQ sequential number
	OemToChar( items[0], msgbuf );
	ListView_SetItemText( hWndList, row, 1, msgbuf );

	// Insert TimeStamp
	OemToChar( items[1], msgbuf );
	ListView_SetItemText( hWndList, row, 2, msgbuf );

	// Insert Delta 
	OemToChar( items[2], msgbuf );
	ListView_SetItemText( hWndList, row, 3, msgbuf );

	return TRUE;
}


/******************************************************************************
*
*	FUNCTION:	UpdateStatistics
*
*	PURPOSE:	Clear the statistics window and refill it with the current 
*				contents of the statistics buffer.  Does not refresh the 
*				buffer from the device driver.
*
******************************************************************************/
void UpdateStatistics( HWND hWnd, HWND hWndList, BOOL Clear )
{
	auto     unsigned long	*ptr;
	int		 totitems, i;
	static   unsigned long seq = 0;		  // Sequential irq #
	static   unsigned long Previous = 0;  // Previous irq timestamp
	TCHAR    TempBuf[128];

	// Start with empty list
	if ( Clear ) {
		// Start from scratch
		seq = 0;
		// post hourglass icon
		SetCapture(hWnd);
		hSaveCursor = SetCursor(hHourGlass);

		// Don't do this because ListView controls under Win95 have
		// a bug that makes it take sometimes several minutes to 
		// delete all the items
		//	ListView_DeleteAllItems( hWndList );
		totitems = ListView_GetItemCount( hWndList );
		Deleting = TRUE;
		for(i = 0; i < totitems; i++)
			ListView_DeleteItem( hWndList, 0 );
		Deleting = FALSE;

		// Initialization done
		SetCursor( hSaveCursor );
		ReleaseCapture();
	}

/*
    if ( ! DeviceIoControl( VxDHandle, CMD_gettime,
					 NULL, 0, &TimeStamp, sizeof TimeStamp,
					 NULL, NULL) )
	{
		Abort( hWnd, "Couldn't access device driver") ;
		return TRUE;
	}
*/

	// Add all List items from Stats[] data
	for ( ptr = Stats; ptr < Stats+StatsLen; ptr++)  {
	 	// Add to list
		long Current = *ptr&0x0FFFFFFF;
		long Delta   = Current-Previous;

		wsprintf(TempBuf," %d \t %ld \t %ld", *ptr>>28, Current, Delta);
		List_Append( hWndList, seq++, TempBuf );
		Previous = Current;
	}

	// Empty the buffer
	StatsLen = 0;

	// Scroll so newly added items are visible
	if ( Autoscroll ) 
		ListView_EnsureVisible( hWndList, ListView_GetItemCount(hWndList)-1, FALSE ); 
}

/****************************************************************************
* 
*    FUNCTION: CreateListView(HWND)
*
*    PURPOSE:  Creates the statistics list view window and initializes it
*
****************************************************************************/
HWND CreateList( HWND hWndParent )                                     
{
	HWND		hWndList;    	  	// handle to list view window
	RECT		rc;         	  	// rectangle for setting size of window
	LV_COLUMN	lvC;				// list view column structure
	DWORD		j;
	static struct {
		TCHAR *	Label;	// title of column
		DWORD	Width;	// width of column in pixels
		DWORD	Fmt;
	} column[] = {
		{	"#"			,	35		},
		{	"IRQ"	    ,	100		},		
		{	"Time Stamp (ticks)", 150 },
		{	"Delta"		,	75		}
	};

	// Ensure that the common control DLL is loaded.
	InitCommonControls();

	// Get the size and position of the parent window.
	GetClientRect( hWndParent, &rc );

	// Create the list view window
	hWndList = CreateWindowEx( 0L, WC_LISTVIEW, "", 
								WS_VISIBLE | WS_CHILD | WS_BORDER | LVS_REPORT |
								    WS_EX_CLIENTEDGE,	// styles
								0, 0, rc.right - rc.left, rc.bottom - rc.top,
								hWndParent,	(HMENU)ID_LIST, hInst, NULL );
	if ( hWndList == NULL )
		return NULL;

	// Initialize columns
	lvC.mask	= LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
	lvC.fmt		= LVCFMT_LEFT;	// left-align column

	// Add the columns.
	for ( j = 0; j < sizeof column/sizeof column[0]; ++j )  {
		lvC.iSubItem	= j;
		lvC.cx			= column[j].Width;
	 	lvC.pszText		= column[j].Label;
		if ( ListView_InsertColumn( hWndList, j, &lvC ) == -1 )
			return NULL;
	}

	return hWndList;
}

/****************************************************************************
*
*	FUNCTION:	InstallHelperReboot
*
*	PURPOSE:	Installs HELPER.VXD in the registry
*
****************************************************************************/

void InstallHelperAndReboot(HWND hWnd)
{
TCHAR	Path[ 256 ];
HKEY	hKeyVxD, hKeyHelper;

	if (MessageBox(hWnd, "Not all interrupts can be monitored     \n" \
	                     "because HELPER.VxD has not been installed. \n" \
					     "Do you want to install it now \n" \
						 "(This will require restarting Windows) ?",
					     "Interrupt Monitor",
	    MB_YESNO | MB_ICONQUESTION) == IDNO)
		return;

		GetCurrentDirectory( sizeof Path, Path );
		wsprintf( Path+lstrlen(Path), "\\%s", HELPER_NAME );
	    if (CreateFile( Path, 0, 0, NULL,
						OPEN_EXISTING, 
						0,
						NULL ) == INVALID_HANDLE_VALUE )  {
				wsprintf( msgbuf, "%s is not present in current directory.", HELPER_NAME );
				Abort( hWnd, msgbuf );
				return;
			}
   
		if  ((RegOpenKey(HKEY_LOCAL_MACHINE, "System\\CurrentControlSet\\Services\\VxD", 
			   &hKeyVxD) == ERROR_SUCCESS) &&
		     (RegCreateKey(hKeyVxD, "HELPER", &hKeyHelper) == ERROR_SUCCESS) &&
			 (RegSetValueEx(hKeyHelper, "StaticVxD", 0, REG_SZ, Path, 
			 sizeof Path) == ERROR_SUCCESS)
			) 
		{
			RegCloseKey (hKeyHelper);
			RegCloseKey (hKeyVxD);
		}
		else
			{
			Abort( hWnd, "Could not write to the Registry. " );
			return;
		}

//		InitiateSystemShutdown(NULL, "Message", 2, FALSE, TRUE);
		ExitWindowsEx(EWX_FORCE | EWX_REBOOT,	0 ) ; // shutdown operation
}

/****************************************************************************
*
*	FUNCTION:	About
*
*	PURPOSE:	Processes messages for "About" dialog box
*
****************************************************************************/

BOOL APIENTRY About( HWND hDlg, UINT message, UINT wParam, LONG lParam )
{
	switch ( message )  {

	   case WM_INITDIALOG:
		  return TRUE;

	   case WM_COMMAND:              
		  if ( LOWORD( wParam ) == IDOK )	 {
			  EndDialog( hDlg, TRUE );
			  return TRUE;
		  }
		  break;
	}
	return FALSE;   
}
